==> GeneR_2.6.0.tar.gz: GeneR/src/compoSeq.cc <==
/*! \file compoSeq.cc
 *  \brief Complex functions for GeneR_seq class
 *  \version 1
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2006-06-05 19:49:26 antoine>
 *  \author A. Lucas from epissage group at CGM codes.
 *  \note Licence: CeCIll
 */

#include <stdio.h>
#include <string>
#include "GeneR_glob.h"
#include "libStrings.h"
#include "compoSeq.h"
#include <math.h>
#include "strcasestr.h"


#define min( A , B )  ( ( A ) < ( B ) ? ( A ) : ( B ) )


extern "C"
{
  void sys_exactmatch (int *bufno,char **word, int *vfrom, int *vto,
		       int *comp, int *wNbOcc, int *step, int *res, 
		       int *nbOcc,int * case_sensible,int *overlap,
		       int *err);
  void composeq (int * bufno, int * from, int * to,int * strand,
		 int * nbregions, int * nbstrand, int * wsize, 
		 char ** labels, int * res,int * step,
		 int * casse, int * err);
  void strcomposeq (char ** seq, int * res, int * wsize, char ** labels, 
		    int * computelabels,int * step,int * casse,int * err);
}


/* ---------------------------------------------------- */
/*                      exactmatch                      */
/* ---------------------------------------------------- */
/** @brief renvoie la position de word dans
           la sequence situe dans le buffer bufno
    @param *bufno  Numero de la sequence
    @param **word Motif a chercher
    @param *vfrom  Positions de debut
    @param *vto    Positions de fin
    @param *comp   0 : brin sens. 1 : brin anti-sens
    @param *wNbOcc Nombre d'occurences desirees
    @param *step   Pas selon lequel le balayage des regions se fait
    @param *res    Tableau de resultats. Positions des occurences
    @param *nbOcc  Nombre d'occurences trouvees. Valeur de retour dans R
    @param *case_sensible whether it takes into account case of letters
    @param *overlap whether we take into account overlapping words
    @param *err    Flag d'erreur
*/
/* ---------------------------------------------------- */

void sys_exactmatch (int *bufno,char **word, int *vfrom, int *vto,
		     int *comp, int *wNbOcc, int *step, int *res, 
		     int *nbOcc,int * case_sensible,int *overlap,
		     int *err)
{
  char *Ptr;
  int lengthWord = strlen (*word);
  int lengthRegion;
  int compt=0;
  int i,j,k;
  int end;

  char * (* fun) (const char*, const char *) = NULL;

  char * posFind;

  char (*upperornot)(char) = NULL;

  fun = strstr;
  upperornot = compoSeq::non_upper;

  if(case_sensible[0]==0)
    fun = strcasestr;
  upperornot = compoSeq::to_upper;


  Ptr = myGeneRbuffer(*bufno,*comp);
  
  if (Ptr == NULL || (! *word))
    {
      *err = -1;
      return;
    }

  //  if(case_sensible[0]==0)
    //    libStrings::sys_upper_string(*word);


  if (*step > 0)
    {
      lengthRegion = (*vto) - (*vfrom) +1;
      
      i= (*vfrom)-1;
      end =  (*vto)-lengthWord;
      while(i<= end && (*nbOcc)<(*wNbOcc) )

	{

	  /*  Ptr+i = &(Ptr[i]) */
	  posFind = fun(Ptr +i, *word);

	  /* when NULL: no motif found */
	  if(posFind == NULL)
	    break;

	  i = (int) (posFind - Ptr) +1;

	  /* check phase */
	  if( (( i - (*vfrom)  ) % *step) != 0)
	      continue;

	  /* here, we found word, and in the correct phase ! */
	  res[*nbOcc] = i;
	  (*nbOcc)++;
	  
	  /* when we DO NOT want overlap */
	  if(*overlap == 0)
	      i+=lengthWord-1;
	      
	}
    }
  else
    { /* reverse*/
      lengthRegion = *vto - *vfrom + 1;

      for(i=(*vto)-lengthWord+1-abs(*step); i>=*vfrom-1 && (*nbOcc)<(*wNbOcc); i=k+(*step))
	{
	  k=i;
	  compt=0;
	  j=0;
	   do {
	     if (upperornot(Ptr[i])==((*word)[j]))
		compt++;
	      i++;j++;
	   } while((upperornot(Ptr[i])==((*word)[j])) && j<lengthWord
		   && i<=(*vto));
	  
	  
	  if (compt==lengthWord)
	    {
	      res[*nbOcc] = i - lengthWord + 1;
	      (*nbOcc)++;

	      /* when we DO NOT want overlap */
	      if(*overlap == 0)
		{
		  k -= ((lengthWord-1) / (*step) ) * (*step) ;
		}
	    }
	}
      
    }

  return;

}



/** 
 * \brief  Compute composition of poly-nucleotides (mono-di-tri... depends on 
 *         wsize)
 * \note   Dans R: verification du buffer (eventuellement du buffer reverse)
 *         verification que length(to) = length(from)
 *         Allocation de res et labels
 *         *nbregions = length(from)
 *         *nbstrand  = length(strand)
 * \param  bufno sequence buffer
 * \param  from,to,strand vector of from, to and strand
 * \param  nbregions size of vectors from, to and strand
 * \param  nbstrand = 1 or nbregions
 * \param  wsize size of partern (wsize=1: count mononucleotides)
 * \param  labels: vector of character of all patern to fill
 * \param  res vector: values will be stored in this variable
 * \param  step Step increment of positions when counting words.
 * \param  casse: 1 (all), 2 (upper) or 3 (lower)
 * \param  err flag to return to R (1=good, 0=error)
 */
void composeq (int * bufno, int * from, int * to,int * strand,
	       int * nbregions, int * nbstrand, int * wsize, 
	       char ** labels, int * res,int * step,
	       int * casse, int * err)
{

  /*
   * longbufno = length of sequence in buffer bufno
   * longvector= length of a line of output matrix
   *           = 5^(*wsize)
   * computelabel: flag (0, or 1) sent to strcomposeq to compute
   *               or not labels.
   * sizealloc : size of *nucleotide = max (to-from +1)
   */
  int i,sizealloc=0,longbufno,onestrand=1,longueur,longvector,computelabel=1;

  char * nucleotides ; /* a temorary  copy of part of buffer */
  int * restmp; /* a temorary  copy res */
  char * Buffie;

  Buffie = myGeneRbuffer(* bufno,0);
  if(!Buffie)
    {
      *err=0;
      return;
    }
  longbufno =   strlen(Buffie);
  longvector = (int) pow (5,(*wsize));
  
  if( (*nbstrand)==(*nbregions))
    onestrand = *nbregions+1;

  /* allocation of temporary sequences alloc = max(to - from +2) */
  for (i=0; i < *nbregions; i++)
    {
      if(to[i] == 0)
	to[i] = longbufno;
      if((to[i] - from[i] + 1) >sizealloc )
	sizealloc =to[i] - from[i] + 1;
    }

  nucleotides = (char *) malloc((sizealloc+1) * sizeof(char));
  if(!nucleotides)
    {
      printf("GeneR.so cannot allocate in composeq");
      *err=0;
      return;
    }

  
  for (i=0; i < *nbregions; i++)
    {
      longueur = to[i]-from[i]+1;
      if(strand[i%onestrand]==1)
	{
	  Buffie = myGeneRbuffer(* bufno,1);
	  from[i]= min(from[i], to[i]) ;
	}
      else
	Buffie = myGeneRbuffer(* bufno,0);



      restmp = res + (i * longvector) ;
      if(longueur<0)
	longueur=1;

      nucleotides = strncpy(nucleotides,Buffie+from[i]-1,longueur);
      nucleotides[longueur] = '\0';
  
      strcomposeq(&nucleotides, restmp,wsize,
		  labels, &computelabel,step,casse,err);
      computelabel = 0; 
    }


  free(nucleotides);

}


namespace compoSeq {
  /** non_upper
   * this function does nothing 
   * \brief aim of this function is to subsitute function to_upper or 
   *  inv_upper for some computations
   * \param c a character
   */
  char non_upper (char c)
  {
    return(c);
  }

  /** to_upper 
   * \brief change char C in upper case
   * \param c character
   */
  char to_upper (char c) 
  {
    if(((c) >= 'a') && ((c) <= 'z'))
      c = c - 'a' + 'A';
    return(c);
  }
  /** inv_upper
   * \brief put in lower case character in upper case and vice versa 
   *\param c character
   */
  char inv_upper (char c) 
  {
    if(((c) >= 'a') && ((c) <= 'z'))
      c = c - 'a' + 'A';
    else
      c = c + 'a' - 'A';
    return(c);
  }
}



/** 
 * function strcomposeq 
 * \brief gives composition in n-nucleotides (mono,bi,tri,etc...)
 *
 * \param  seq sequence 
 * \param  wsize size of partern (wsize=1: count mononucleotides)
 * \param  labels: vector of character of all patern to fill
 * \param  res vector: values will be stored in this variable
 * \param  computelabels flag if 1 we compute labels
 * \param  step Step increment of positions when counting words.
 * \param  casse 1 (all), 2 (upper) or 3 (lower)
 * \param  err flag to return to R (1=good, 0=error) 
 * \note this function can be called by composeq 
 */
void strcomposeq (char ** seq, int * res, int * wsize, char ** labels, 
		  int * computelabels,int * step,int * casse,int * err)
{
  int i,k,j,l,m;
  int seqsize;
  int indice;
  char mot[6]="TCAGX";
  /*  char * copyseq; */
  char * tmplabel;

  char (*upperorlower)(char) = NULL;

  upperorlower = compoSeq::to_upper;
  /* casse vaux 1 (all), 2 (upper) ou 3 (lower)*/
  if(casse[0]==2)
      upperorlower = compoSeq::non_upper;

  if(casse[0]==3)
      upperorlower = compoSeq::inv_upper;

  seqsize = strlen( *seq) ;
  k=l=0;

  /* counting */
  // for(j = 0; j <= ((*wsize - 1))*(*allphases); j++)
    
  /* We put copyseq on phase j */
  j=0;
  /* printf("j %d, s %s\n",j,copyseq); */
  while (j < (1+seqsize - (*wsize) ) )
    {
      indice =0;
      for ( k=(*wsize)-1; k>=0; k--)
	{
	  switch(upperorlower((*seq)[j++]))
	    {
	    case 'T': break;
	    case 'C': indice = indice + (int)pow(5,k); break;
	    case 'A': indice = indice + 2* (int)pow(5,k); break;
	    case 'G': indice = indice + 3*(int)pow(5,k); break;
	    default:  indice = indice + 4*(int)pow(5,k); 
	    }
	}
      j = j - (*wsize) + *step;
      (res[indice])++;
    }
   

  /* Generating labels 
   *
   * Example *wize = 3
   * i = a*5^2 + b*5 + c
   * then tmplabel will be concatenation( mot[a] mot[b] mot[c] )
   * with mot = "TCAGX", a,b,c in [[0,4]]
   */
  if(*computelabels)
    for (i=0; i< pow(5, (*wsize)); i++)
      {
	m=0;
	l = i;
	
	tmplabel = labels[i];
	for ( k=(*wsize)-1; k>=0; k--)
	  {
	    m = l % 5;
	    tmplabel[k] = mot[m];
	    l = (l-m)/ 5;
	  }
	tmplabel[(*wsize)]='\0';
      }
}

==> GeneR_2.10.0.tar.gz: GeneR/src/compoSeq.cc <==
/*! \file compoSeq.cc
 *  \brief Complex functions for GeneR_seq class
 *  \version 1
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2007-10-12 18:36:10 antoine>
 *  \author A. Lucas from epissage group at CGM codes.
 *  \note Licence: CeCIll
 */

#include <stdio.h>
#include <string>
#include "GeneR_glob.h"
#include "libStrings.h"
#include "compoSeq.h"
#include <math.h>
#include "strcasestr.h"


#define min( A , B )  ( ( A ) < ( B ) ? ( A ) : ( B ) )


extern "C"
{
  void sys_exactmatch (int *bufno,char **word, int *vfrom, int *vto,
		       int *comp, int *wNbOcc, int *step, int *res, 
		       int *nbOcc,int * case_sensible,int *overlap,
		       int *err);
  void composeq (int * bufno, int * from, int * to,int * strand,
		 int * nbregions, int * nbstrand, int * wsize, 
		 char ** labels, int * res,int * step,
		 int * casse, int * err);
  void strcomposeq (char ** seq, int * res, int * wsize, char ** labels, 
		    int * computelabels,int * step,int * casse,int * err);
}


/* ---------------------------------------------------- */
/*                      exactmatch                      */
/* ---------------------------------------------------- */
/** @brief renvoie la position de word dans
           la sequence situe dans le buffer bufno
    @param *bufno  Numero de la sequence
    @param **word Motif a chercher
    @param *vfrom  Positions de debut
    @param *vto    Positions de fin
    @param *comp   0 : brin sens. 1 : brin anti-sens
    @param *wNbOcc Nombre d'occurences desirees
    @param *step   Pas selon lequel le balayage des regions se fait
    @param *res    Tableau de resultats. Positions des occurences
    @param *nbOcc  Nombre d'occurences trouvees. Valeur de retour dans R
    @param *case_sensible whether it takes into account case of letters
    @param *overlap whether we take into account overlapping words
    @param *err    Flag d'erreur
*/
/* ---------------------------------------------------- */

void sys_exactmatch (int *bufno,char **word, int *vfrom, int *vto,
		     int *comp, int *wNbOcc, int *step, int *res, 
		     int *nbOcc,int * case_sensible,int *overlap,
		     int *err)
{
  char *Ptr;
  int lengthWord = strlen (*word);
  int lengthRegion;
  int compt=0;
  int i,j,k;
  int end;

  char * (* fun) (const char*, const char *) = NULL;

  char * posFind;

  char (*upperornot)(char) = NULL;

  fun = strstr;
  upperornot = compoSeq::non_upper;

  if(case_sensible[0]==0)
    fun = strcasestr;
  upperornot = compoSeq::to_upper;


  Ptr = GeneR_glob::instance()->buffer(*bufno,*comp);
  
  if (Ptr == NULL || (! *word))
    {
      *err = -1;
      return;
    }

  //  if(case_sensible[0]==0)
    //    libStrings::sys_upper_string(*word);


  if (*step > 0)
    {
      lengthRegion = (*vto) - (*vfrom) +1;
      
      i= (*vfrom)-1;
      end =  (*vto)-lengthWord;
      while(i<= end && (*nbOcc)<(*wNbOcc) )

	{

	  /*  Ptr+i = &(Ptr[i]) */
	  posFind = fun(Ptr +i, *word);

	  /* when NULL: no motif found */
	  if(posFind == NULL)
	    break;

	  i = (int) (posFind - Ptr) +1;

	  /* check phase */
	  if( (( i - (*vfrom)  ) % *step) != 0)
	      continue;

	  /* here, we found word, and in the correct phase ! */
	  res[*nbOcc] = i;
	  (*nbOcc)++;
	  
	  /* when we DO NOT want overlap */
	  if(*overlap == 0)
	      i+=lengthWord-1;
	      
	}
    }
  else
    { /* reverse*/
      lengthRegion = *vto - *vfrom + 1;

      for(i=(*vto)-lengthWord+1-abs(*step); i>=*vfrom-1 && (*nbOcc)<(*wNbOcc); i=k+(*step))
	{
	  k=i;
	  compt=0;
	  j=0;
	   do {
	     if (upperornot(Ptr[i])==((*word)[j]))
		compt++;
	      i++;j++;
	   } while((upperornot(Ptr[i])==((*word)[j])) && j<lengthWord
		   && i<=(*vto));
	  
	  
	  if (compt==lengthWord)
	    {
	      res[*nbOcc] = i - lengthWord + 1;
	      (*nbOcc)++;

	      /* when we DO NOT want overlap */
	      if(*overlap == 0)
		{
		  k -= ((lengthWord-1) / (*step) ) * (*step) ;
		}
	    }
	}
      
    }

  return;

}



/** 
 * \brief  Compute composition of poly-nucleotides (mono-di-tri... depends on 
 *         wsize)
 * \note   Dans R: verification du buffer (eventuellement du buffer reverse)
 *         verification que length(to) = length(from)
 *         Allocation de res et labels
 *         *nbregions = length(from)
 *         *nbstrand  = length(strand)
 * \param  bufno sequence buffer
 * \param  from,to,strand vector of from, to and strand
 * \param  nbregions size of vectors from, to and strand
 * \param  nbstrand = 1 or nbregions
 * \param  wsize size of partern (wsize=1: count mononucleotides)
 * \param  labels: vector of character of all patern to fill
 * \param  res vector: values will be stored in this variable
 * \param  step Step increment of positions when counting words.
 * \param  casse: 1 (all), 2 (upper) or 3 (lower)
 * \param  err flag to return to R (1=good, 0=error)
 */
void composeq (int * bufno, int * from, int * to,int * strand,
	       int * nbregions, int * nbstrand, int * wsize, 
	       char ** labels, int * res,int * step,
	       int * casse, int * err)
{

  /*
   * longbufno = length of sequence in buffer bufno
   * longvector= length of a line of output matrix
   *           = 5^(*wsize)
   * computelabel: flag (0, or 1) sent to strcomposeq to compute
   *               or not labels.
   * sizealloc : size of *nucleotide = max (to-from +1)
   */
  int i,sizealloc=0,longbufno,onestrand=1,longueur,longvector,computelabel=1;

  char * nucleotides ; /* a temorary  copy of part of buffer */
  int * restmp; /* a temorary  copy res */
  char * Buffie;

  Buffie = GeneR_glob::instance()->buffer(* bufno,0);
  if(!Buffie)
    {
      *err=0;
      return;
    }
  longbufno =   strlen(Buffie);
  longvector = (int) pow (5,(*wsize));
  
  if( (*nbstrand)==(*nbregions))
    onestrand = *nbregions+1;

  /* allocation of temporary sequences alloc = max(to - from +2) */
  for (i=0; i < *nbregions; i++)
    {
      if(to[i] == 0)
	to[i] = longbufno;
      if((to[i] - from[i] + 1) >sizealloc )
	sizealloc =to[i] - from[i] + 1;
    }

  nucleotides = (char *) malloc((sizealloc+1) * sizeof(char));
  if(!nucleotides)
    {
      printf("GeneR.so cannot allocate in composeq");
      *err=0;
      return;
    }

  
  for (i=0; i < *nbregions; i++)
    {
      longueur = to[i]-from[i]+1;
      if(strand[i%onestrand]==1)
	{
	  Buffie = GeneR_glob::instance()->buffer(* bufno,1);
	  from[i]= min(from[i], to[i]) ;
	}
      else
	Buffie = GeneR_glob::instance()->buffer(* bufno,0);



      restmp = res + (i * longvector) ;
      if(longueur<0)
	longueur=1;

      nucleotides = strncpy(nucleotides,Buffie+from[i]-1,longueur);
      nucleotides[longueur] = '\0';
  
      strcomposeq(&nucleotides, restmp,wsize,
		  labels, &computelabel,step,casse,err);
      computelabel = 0; 
    }


  free(nucleotides);

}


namespace compoSeq {
  /** non_upper
   * this function does nothing 
   * \brief aim of this function is to subsitute function to_upper or 
   *  inv_upper for some computations
   * \param c a character
   */
  char non_upper (char c)
  {
    return(c);
  }

  /** to_upper 
   * \brief change char C in upper case
   * \param c character
   */
  char to_upper (char c) 
  {
    if(((c) >= 'a') && ((c) <= 'z'))
      c = c - 'a' + 'A';
    return(c);
  }
  /** inv_upper
   * \brief put in lower case character in upper case and vice versa 
   *\param c character
   */
  char inv_upper (char c) 
  {
    if(((c) >= 'a') && ((c) <= 'z'))
      c = c - 'a' + 'A';
    else
      c = c + 'a' - 'A';
    return(c);
  }
}



/** 
 * function strcomposeq 
 * \brief gives composition in n-nucleotides (mono,bi,tri,etc...)
 *
 * \param  seq sequence 
 * \param  wsize size of partern (wsize=1: count mononucleotides)
 * \param  labels: vector of character of all patern to fill
 * \param  res vector: values will be stored in this variable
 * \param  computelabels flag if 1 we compute labels
 * \param  step Step increment of positions when counting words.
 * \param  casse 1 (all), 2 (upper) or 3 (lower)
 * \param  err flag to return to R (1=good, 0=error) 
 * \note this function can be called by composeq 
 */
void strcomposeq (char ** seq, int * res, int * wsize, char ** labels, 
		  int * computelabels,int * step,int * casse,int * err)
{
  int i,k,j,l,m;
  int seqsize;
  int indice;
  char mot[6]="TCAGX";
  /*  char * copyseq; */
  char * tmplabel;

  char (*upperorlower)(char) = NULL;

  upperorlower = compoSeq::to_upper;
  /* casse vaux 1 (all), 2 (upper) ou 3 (lower)*/
  if(casse[0]==2)
      upperorlower = compoSeq::non_upper;

  if(casse[0]==3)
      upperorlower = compoSeq::inv_upper;

  seqsize = strlen( *seq) ;
  k=l=0;

  /* counting */
  // for(j = 0; j <= ((*wsize - 1))*(*allphases); j++)
    
  /* We put copyseq on phase j */
  j=0;
  /* printf("j %d, s %s\n",j,copyseq); */
  while (j < (1+seqsize - (*wsize) ) )
    {
      indice =0;
      for ( k=(*wsize)-1; k>=0; k--)
	{
	  switch(upperorlower((*seq)[j++]))
	    {
	    case 'T': break;
	    case 'C': indice = indice + (int)pow(5,k); break;
	    case 'A': indice = indice + 2* (int)pow(5,k); break;
	    case 'G': indice = indice + 3*(int)pow(5,k); break;
	    default:  indice = indice + 4*(int)pow(5,k); 
	    }
	}
      j = j - (*wsize) + *step;
      (res[indice])++;
    }
   

  /* Generating labels 
   *
   * Example *wize = 3
   * i = a*5^2 + b*5 + c
   * then tmplabel will be concatenation( mot[a] mot[b] mot[c] )
   * with mot = "TCAGX", a,b,c in [[0,4]]
   */
  if(*computelabels)
    for (i=0; i< pow(5, (*wsize)); i++)
      {
	m=0;
	l = i;
	
	tmplabel = labels[i];
	for ( k=(*wsize)-1; k>=0; k--)
	  {
	    m = l % 5;
	    tmplabel[k] = mot[m];
	    l = (l-m)/ 5;
	  }
	tmplabel[(*wsize)]='\0';
      }
}
